# generated by datamodel-codegen:
#   filename:  openapi.yaml
#   timestamp: 2025-10-08T12:50:40+00:00

from __future__ import annotations

from datetime import date
from enum import Enum
from typing import Annotated, Any, Optional, Union

from pydantic import AnyUrl, AwareDatetime, BaseModel, ConfigDict, Field, RootModel


class CRS(Enum):
    http___www_opengis_net_def_crs_OGC_1_3_CRS84 = (
        "http://www.opengis.net/def/crs/OGC/1.3/CRS84"
    )
    http___www_opengis_net_def_crs_OGC_0_CRS84h = (
        "http://www.opengis.net/def/crs/OGC/0/CRS84h"
    )


class ApiError(BaseModel):
    """
    API error based on RFC 7807
    """

    model_config = ConfigDict(
        extra="allow",
    )
    type: str
    title: Optional[str] = None
    status: Optional[int] = None
    detail: Optional[str] = None
    instance: Optional[str] = None
    traceback: Optional[list[str]] = None


class MaxOccurs(Enum):
    unbounded = "unbounded"


class Link(BaseModel):
    href: str
    rel: Optional[str] = Field(None, examples=["service"])
    type: Optional[str] = Field(None, examples=["application/json"])
    hreflang: Optional[str] = Field(None, examples=["en"])
    title: Optional[str] = None


class Bbox(BaseModel):
    bbox: list[float] = Field(..., max_length=4, min_length=4)
    crs: Optional[CRS] = CRS.http___www_opengis_net_def_crs_OGC_1_3_CRS84


class DataType(Enum):
    array = "array"
    boolean = "boolean"
    integer = "integer"
    number = "number"
    object = "object"
    string = "string"


class Metadata(BaseModel):
    title: Optional[str] = None
    role: Optional[str] = None
    href: Optional[str] = None


class AdditionalParameter(BaseModel):
    name: str
    value: list[Union[str, float, int, list[dict[str, Any]], dict[str, Any]]]


class TransmissionMode(Enum):
    value = "value"
    reference = "reference"


class Capabilities(BaseModel):
    title: Optional[str] = Field(None, examples=["Example processing server"])
    description: Optional[str] = Field(
        None,
        examples=["Example server implementing the OGC API - Processes 1.0 Standard"],
    )
    links: list[Link]


class ConformanceDeclaration(BaseModel):
    conformsTo: list[str]


class ResponseType(Enum):
    raw = "raw"
    document = "document"


class Subscriber(BaseModel):
    """
    Optional URIs for callbacks for this job.

    Support for this parameter is not required and the parameter may be
    removed from the API definition, if conformance class **'callback'**
    is not listed in the conformance declaration under `/conformance`.
    """

    successUri: Optional[AnyUrl] = None
    inProgressUri: Optional[AnyUrl] = None
    failedUri: Optional[AnyUrl] = None


class JobStatus(Enum):
    accepted = "accepted"
    running = "running"
    successful = "successful"
    failed = "failed"
    dismissed = "dismissed"


class JobType(Enum):
    process = "process"


class JobControlOptions(Enum):
    sync_execute = "sync-execute"
    async_execute = "async-execute"
    dismiss = "dismiss"


class JobInfo(BaseModel):
    processID: Optional[str] = None
    type: JobType
    jobID: str
    status: JobStatus
    message: Optional[str] = None
    created: Optional[AwareDatetime] = None
    started: Optional[AwareDatetime] = None
    finished: Optional[AwareDatetime] = None
    updated: Optional[AwareDatetime] = None
    progress: Annotated[Optional[int], Field(None, ge=0, le=100)] = None
    links: Optional[list[Link]] = None
    traceback: Optional[list[str]] = None


class InlineValue(
    RootModel[
        Union[
            Any,
            bool,
            bytes,
            AnyUrl,
            date,
            AwareDatetime,
            str,
            int,
            float,
            list,
            dict[str, Any],
            Bbox,
        ]
    ]
):
    root: Union[
        Any,
        bool,
        bytes,
        AnyUrl,
        date,
        AwareDatetime,
        str,
        int,
        float,
        list,
        dict[str, Any],
        Bbox,
    ]


class AdditionalParameters(Metadata):
    parameters: Optional[list[AdditionalParameter]] = None


class DescriptionType(BaseModel):
    title: Optional[str] = None
    description: Optional[str] = None
    keywords: Optional[list[str]] = None
    metadata: Optional[list[Metadata]] = None
    additionalParameters: Optional[AdditionalParameters] = None


class ProcessSummary(DescriptionType):
    id: str
    version: str
    jobControlOptions: Optional[list[JobControlOptions]] = None
    outputTransmission: Optional[list[TransmissionMode]] = None
    links: Optional[list[Link]] = None


class JobList(BaseModel):
    jobs: list[JobInfo]
    links: list[Link]


class ProcessList(BaseModel):
    processes: list[ProcessSummary]
    links: list[Link]


class Format(BaseModel):
    mediaType: Optional[str] = None
    encoding: Optional[str] = None
    schema_: Optional[Union[AnyUrl, Schema]] = Field(None, alias="schema")


# TODO - Code has been manually moved here after class Format,
#  because it was wrongly placed last by the datamodel-code-generator tool
#  That's why we also use the tool with keep_model_order=False - which is by
#  the way also a bug, as it does the opposite of what it should.
#  See https://github.com/koxudaxi/datamodel-code-generator/issues/2508
#  PLEASE MAKE SURE YOU INCLUDE THIS COMMENT AS LONG AS THE PROBLEM PERSISTS.
class QualifiedValue(Format):
    value: InlineValue


class InlineOrRefValue(RootModel[Union[InlineValue, QualifiedValue, Link]]):
    root: Union[InlineValue, QualifiedValue, Link]


class Discriminator(BaseModel):
    propertyName: Optional[str] = Field(None, min_length=1)
    mapping: Optional[dict[str, Schema]] = None


class Schema(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
    )
    field_ref: Optional[str] = Field(None, alias="$ref")
    title: Optional[str] = None
    multipleOf: Optional[float] = Field(None, gt=0.0)
    maximum: Optional[float] = None
    exclusiveMaximum: Optional[bool] = False
    minimum: Optional[float] = None
    exclusiveMinimum: Optional[bool] = False
    maxLength: Optional[int] = Field(None, ge=0)
    minLength: Optional[int] = Field(0, ge=0)
    pattern: Optional[str] = None
    maxItems: Optional[int] = Field(None, ge=0)
    minItems: Optional[int] = Field(0, ge=0)
    uniqueItems: Optional[bool] = False
    maxProperties: Optional[int] = Field(None, ge=0)
    minProperties: Optional[int] = Field(0, ge=0)
    required: Optional[list[str]] = Field(None, min_length=1)
    enum: Optional[list] = Field(None, min_length=1)
    type: Optional[DataType] = None
    not_: Optional[Schema] = Field(None, alias="not")
    allOf: Optional[list[Schema]] = None
    oneOf: Optional[list[Schema]] = None
    anyOf: Optional[list[Schema]] = None
    discriminator: Optional[Discriminator] = None
    items: Optional[Union[list[Schema], Schema]] = None
    properties: Optional[dict[str, Schema]] = None
    additionalProperties: Optional[Union[Schema, bool]] = True
    description: Optional[str] = None
    format: Optional[str] = None
    default: Optional[Any] = None
    nullable: Optional[bool] = False
    readOnly: Optional[bool] = False
    writeOnly: Optional[bool] = False
    example: Optional[Any] = None
    examples: Optional[Any] = None
    deprecated: Optional[bool] = False
    contentMediaType: Optional[str] = None
    contentEncoding: Optional[str] = None
    contentSchema: Optional[str] = None


class ProcessDescription(ProcessSummary):
    inputs: Optional[dict[str, InputDescription]] = None
    outputs: Optional[dict[str, OutputDescription]] = None


class ProcessRequest(BaseModel):
    inputs: Optional[dict[str, Any]] = None
    outputs: Optional[dict[str, Output]] = None
    response: Optional[ResponseType] = ResponseType.raw
    subscriber: Optional[Subscriber] = None


class InputDescription(DescriptionType):
    minOccurs: Optional[int] = 1
    maxOccurs: Optional[Union[int, MaxOccurs]] = None
    schema_: Schema = Field(..., alias="schema")


class OutputDescription(DescriptionType):
    schema_: Schema = Field(..., alias="schema")


class Output(BaseModel):
    format: Optional[Format] = None
    transmissionMode: Optional[TransmissionMode] = TransmissionMode.value


class JobResults(RootModel[Optional[dict[str, InlineOrRefValue]]]):
    root: Optional[dict[str, InlineOrRefValue]] = None


Format.model_rebuild()
InlineOrRefValue.model_rebuild()
Discriminator.model_rebuild()
ProcessDescription.model_rebuild()
ProcessRequest.model_rebuild()
QualifiedValue.model_rebuild()



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import importlib
from typing import Optional, TypeVar

_type_fn = type

T = TypeVar("T")


# noinspection PyShadowingBuiltins
def import_value(
    ref: str,
    *,
    type: type[T],
    name: str,
    example: Optional[str] = None,
) -> T:
    """
    Dynamically import a value given by its reference `ref`.

    Args:
        ref: A value reference of the form "<module>:<attr>".
        type: Expected type of the reference
        name: Name of the reference
        example: An example reference (for error messages)

    Returns:
        The imported value.

    Raises:
        ValueError: if the `ref` syntax is invalid, or the module
            cannot be imported, or attribute cannot be found.
        TypeError: if the loaded value is not an instance of
            the expected type.
    """
    example = example or "path.to.module:attr"

    try:
        module_name, attr_ref = ref.split(":", maxsplit=1)
    except ValueError:
        raise ValueError(
            f"The {name} reference must be passed in the form {example!r}, "
            f"but got {ref!r}."
        )

    try:
        module = importlib.import_module(module_name)
    except ImportError:
        raise ValueError(f"Cannot import module {module_name!r}.")

    value = module
    partial_spec = module_name
    for i, attr_name in enumerate(attr_ref.split(".")):
        try:
            value = getattr(value, attr_name)
        except AttributeError:
            raise ValueError(
                f"{'Module' if i == 0 else 'Object'} {partial_spec!r} "
                f"has no attribute {attr_name!r}."
            )
        partial_spec += (":" if i == 0 else ".") + attr_name

    if not isinstance(value, type):
        raise TypeError(
            f"The reference {ref!r} should refer to a {name} of type "
            f"{type.__name__}, but got type {_type_fn(value).__name__}."
        )

    return value



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import sys
from io import StringIO
from pathlib import Path
from typing import Annotated, Any

import click
import pydantic
from pydantic import Field

from gavicore.models import ProcessDescription, ProcessRequest
from .obj import flatten_obj, nest_dict

SUBSCRIBER_EVENTS = {
    "success": "successUri",
    "failed": "failedUri",
    "progress": "inProgressUri",
}


# noinspection PyShadowingBuiltins
class ExecutionRequest(ProcessRequest):
    """
    Process execution request.
    Extends [ProcessRequest][ProcessRequest]

    - to allow the process identifier being part of the request,
    - to allow creating nested object values for input names with dots.

    Args:
        process_id: Process identifier
        dotpath: Whether dots in input names should be used to create
            nested object values. Defaults to `False`.
        inputs: Optional process inputs given as key-value mapping.
            Values may be of any JSON-serializable type accepted by
            the given process.
        outputs: Optional process outputs given as key-value mapping.
            Values are of type [Output][procodile.models.Output]
            supported by the given process.
        subscriber: Optional subscriber of type
            [Subscriber][procodile.models.Subscriber] comprising callback
            URLs that are informed about process status changes
            while the processing takes place.
    """

    process_id: Annotated[str, Field(title="Process identifier", min_length=1)]
    dotpath: Annotated[
        bool, Field(title="Whether to encode nested input values using dots ('.').")
    ] = False

    def to_process_request(self) -> ProcessRequest:
        """
        Convert this execution request into a process request as used by the
        `execute-process` operation.
        """
        inputs = self.inputs
        if inputs and self.dotpath:
            inputs = nest_dict(inputs)
        return ProcessRequest(
            inputs=inputs,
            outputs=self.outputs,
            response=self.response,
            subscriber=self.subscriber,
        )

    @classmethod
    def create(
        cls,
        process_id: str | None = None,
        dotpath: bool = False,
        request_path: str | None = None,
        inputs: list[str] | None = None,
        subscribers: list[str] | None = None,
    ) -> "ExecutionRequest":
        request_dict, _ = _read_execution_request(request_path)
        if process_id:
            request_dict["process_id"] = process_id
        if dotpath:
            request_dict["dotpath"] = dotpath
        inputs_dict = _parse_inputs(inputs)
        if inputs_dict:
            request_dict["inputs"] = dict(request_dict.get("inputs") or {})
            request_dict["inputs"].update(inputs_dict)
        subscriber_dict = _parse_subscribers(subscribers)
        if subscriber_dict:
            request_dict["subscriber"] = dict(request_dict.get("subscriber") or {})
            request_dict["subscriber"].update(subscriber_dict)
        try:
            return ExecutionRequest(**request_dict)
        except pydantic.ValidationError as e:
            raise click.ClickException(f"Execution request is invalid: {e}")

    @classmethod
    def from_process_description(
        cls,
        process_description: ProcessDescription,
        dotpath: bool = False,
    ) -> "ExecutionRequest":
        """
        Create an execution request from the given process description.

        Args:
            process_description: The process description
            dotpath: Whether to allow for dot-separated input
                names for nested object values

        Returns:
            The execution requests populated with default values.
        """
        return _from_process_description(process_description, dotpath)


def _read_execution_request(
    request_path: Path | str | None = None,
) -> tuple[dict[str, Any], str]:
    if not request_path:
        return {}, ""

    if request_path == "-":
        request_path = "<stdin>"
        # content = "\n".join(line for line in sys.stdin).strip()
        content = sys.stdin.read().strip()
        file_format = "json" if content.startswith("{") else "yaml"
    else:
        path = Path(request_path)
        content = path.read_text()
        file_format = "json" if path.suffix in (".json", ".JSON") else "yaml"

    if file_format == "json":
        import json

        request_dict = json.loads(content)
    else:
        import yaml

        request_dict = yaml.safe_load(StringIO(content))

    if not isinstance(request_dict, dict):
        raise click.ClickException(
            f"Request must be an object, but was type {type(request_dict).__name__}"
        )

    return request_dict, str(request_path)


def _parse_inputs(inputs: list[str] | None) -> dict[str, Any]:
    return dict(_parse_inputs_kv(kv) for kv in (inputs or []))


def _parse_inputs_kv(kv: str) -> tuple[str, str]:
    parts = kv.split("=", maxsplit=1)
    key, value = parts if len(parts) == 2 else (parts[0], "true")
    return _parse_input_name(key), _parse_input_value(value)


def _parse_input_name(key: str) -> str:
    norm_key = key.strip().replace("-", "_")

    property_names = key.split(".")
    for property_name in property_names:
        if not property_name.isidentifier():
            raise click.ClickException(f"Invalid request NAME: {key!r}")

    return norm_key


def _parse_input_value(value: str) -> Any:
    import json

    try:
        return json.loads(value)
    except json.JSONDecodeError:
        return value


def _parse_subscribers(subscribers: list[str] | None) -> dict[str, str]:
    return dict(_parse_subscriber_kv(kv) for kv in (subscribers or []))


def _parse_subscriber_kv(kv: str) -> tuple[str, str]:
    try:
        key, value = kv.split("=", maxsplit=1)
    except ValueError:
        raise click.ClickException(
            f"Invalid subscriber item: must have form `EVENT=URL`, but was {kv!r}"
        )
    return _parse_subscriber_event(key), _parse_subscriber_url(value)


def _parse_subscriber_event(key: str):
    norm_key = SUBSCRIBER_EVENTS.get(key)
    if norm_key is None:
        raise click.ClickException(
            "Invalid subscriber EVENT: must be one of "
            f"[{'|'.join(SUBSCRIBER_EVENTS.keys())}], but was {key!r}"
        )
    return norm_key


def _parse_subscriber_url(value: str):
    from urllib.parse import urlparse

    url = urlparse(value)
    if not all([url.scheme in ("http", "https"), url.netloc]):
        raise click.ClickException(f"Invalid subscriber URL: {value!r}")
    return value


# noinspection PyShadowingBuiltins
def _from_process_description(
    process_description: ProcessDescription, dotpath: bool
) -> ExecutionRequest:
    inputs = {
        k: _get_schema_default_value(
            v.schema_.model_dump(mode="json", exclude_unset=True) if v.schema_ else None
        )
        for k, v in (process_description.inputs or {}).items()
    }
    if dotpath:
        inputs = flatten_obj(inputs)
    return ExecutionRequest(
        process_id=process_description.id, dotpath=dotpath, inputs=inputs
    )


def _get_schema_default_value(schema: Any) -> Any:
    if schema and isinstance(schema, dict):
        if "default" in schema:
            return schema["default"]
        if "enum" in schema and isinstance(schema["enum"], list) and schema["enum"]:
            return schema["enum"][0]
        if "type" in schema and isinstance(schema["type"], str) and schema["type"]:
            type_ = schema["type"]
            if type_ == "object" and "properties" in schema:
                properties_ = schema["properties"]
                if properties_ and isinstance(properties_, dict):
                    return {
                        p_name: _get_schema_default_value(p_schema)
                        for p_name, p_schema in properties_.items()
                    }
            return _JSON_DATA_TYPE_DEFAULT_VALUES.get(type_)
    return None


_JSON_DATA_TYPE_DEFAULT_VALUES: dict[str, Any] = {
    "null": None,
    "boolean": False,
    "integer": 0,
    "number": 0,
    "string": "",
    "array": [],
    "object": {},
}



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import typer

# See also cuiman/src/cuiman/cli/cli.py
PROCESS_ID_ARGUMENT = typer.Argument(
    help="Process identifier.",
)

# See also cuiman/src/cuiman/cli/cli.py
REQUEST_INPUT_OPTION = typer.Option(
    "--input",
    "-i",
    help="Process input value.",
    metavar="[NAME=VALUE]...",
)

# See also cuiman/src/cuiman/cli/cli.py
REQUEST_SUBSCRIBER_OPTION = typer.Option(
    "--subscriber",
    "-s",
    help="Process subscriber URL.",
    metavar="[NAME=URL]...",
)

DOT_PATH_OPTION = typer.Option(
    ...,
    "--dotpath",
    "-d",
    is_flag=True,
    help="Input names use dot-path notion to encode nested values, e.g., `-i scene.colors.bg=red`.",
)

# See also cuiman/src/cuiman/cli/cli.py
REQUEST_FILE_OPTION = typer.Option(
    ...,
    "--request",
    "-r",
    help="Execution request file. Use `-` to read from <stdin>.",
    metavar="PATH",
)



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import copy
import json
from typing import Any

import pydantic

from gavicore.models import Schema


def create_schema_instance(name: str, schema: dict[str, Any]) -> Schema:
    try:
        return Schema(**schema)
    except pydantic.ValidationError:
        print(80 * ">")
        print("name:", name)
        print("schema:", json.dumps(schema, indent=2))
        print(80 * "<")
        raise


def create_json_schema(
    model_class: type[pydantic.BaseModel],
) -> dict[str, Any]:
    model_class.model_rebuild()
    schema = model_class.model_json_schema()
    schema = inline_schema_refs(schema)
    return backport_schema_to_openapi_3_0(schema)


def inline_schema_refs(schema: dict[str, Any]) -> dict[str, Any]:
    defs: dict[str, Any] | None = schema.get("$defs")
    if not defs:
        return schema
    schema = copy.copy(schema)
    schema.pop("$defs")
    return _inline_schema_refs(schema, {f"#/$defs/{k}": v for k, v in defs.items()})


def _inline_schema_refs(schema: dict[str, Any], defs: dict[str, Any]) -> dict[str, Any]:
    if "$ref" in schema:
        ref = schema["$ref"]
        if ref in defs:
            ref_schema = _inline_schema_refs(defs[ref], defs)
            schema = copy.copy(schema)
            schema.pop("$ref")
            schema.update(copy.deepcopy(ref_schema))
            return schema
    schema = copy.copy(schema)
    for k in ("allOf", "anyOf", "oneOf"):
        if k in schema:
            if isinstance(schema[k], list):
                schema[k] = [_inline_schema_refs(s, defs) for s in schema[k]]
    for k in ("items", "prefixItems", "additionalProperties"):
        if k in schema:
            if isinstance(schema[k], list):
                schema[k] = [_inline_schema_refs(s, defs) for s in schema[k]]
            else:
                schema[k] = _inline_schema_refs(schema[k], defs)
    for k in ("properties",):
        if k in schema:
            if isinstance(schema[k], dict):
                schema[k] = {
                    k: _inline_schema_refs(s, defs) for k, s in schema[k].items()
                }
    return schema


def backport_schema_to_openapi_3_0(schema: dict[str, Any]) -> dict[str, Any]:
    if "type" in schema:
        type_ = schema["type"]
        if type_ == "null":
            schema.pop("type")
            schema["nullable"] = True
    if "prefixItems" in schema:
        prefix_items = schema.pop("prefixItems")
        if prefix_items:
            # Care, this is a naive implementation:
            schema["items"] = backport_schema_to_openapi_3_0(prefix_items[0])
    for schema_key in ("items", "additionalProperties"):
        if schema_key in schema:
            if isinstance(schema[schema_key], (list, tuple)):
                schema[schema_key] = list(
                    map(backport_schema_to_openapi_3_0, schema[schema_key])
                )
            elif isinstance(schema[schema_key], dict):
                schema[schema_key] = backport_schema_to_openapi_3_0(schema[schema_key])
    for dict_key in ("properties", "$defs"):
        if dict_key in schema and isinstance(schema[dict_key], dict):
            schema[dict_key] = {
                k: backport_schema_to_openapi_3_0(v)
                for k, v in schema[dict_key].items()
            }
    for x_of_key in ("allOf", "anyOf", "oneOf"):
        if x_of_key in schema:
            schema[x_of_key] = list(
                map(backport_schema_to_openapi_3_0, schema[x_of_key])
            )
            if len(schema[x_of_key]) == 2:
                x_of_copy = [s for s in schema[x_of_key] if s != {"nullable": True}]
                if len(x_of_copy) == 1:
                    schema.pop(x_of_key)
                    schema.update(x_of_copy[0])
                    schema["nullable"] = True
    return schema



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import click
import typer.core


class AliasedGroup(typer.core.TyperGroup):
    """
    A group that accepts command aliases created from the
    first letters of the words after splitting a command name
    at hyphens.
    """

    @staticmethod
    def to_alias(name: str):
        """Create a short alias for given command name."""
        return "".join(map(lambda n: n[0], name.split("-")))

    def get_command(self, ctx, cmd_name):
        rv = super().get_command(ctx, cmd_name)

        if rv is not None:
            return rv

        matches = [
            x
            for x in self.list_commands(ctx)
            if cmd_name == x or cmd_name == self.to_alias(x)
        ]

        if not matches:
            return None

        if len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])

        ctx.fail(f"Too many matches: {', '.join(sorted(matches))}")

    def resolve_command(
        self, ctx, args
    ) -> tuple[str | None, click.Command | None, list[str]]:
        # always return the full command name
        _, cmd, args = super().resolve_command(ctx, args)
        if cmd is not None:
            return cmd.name, cmd, args
        else:
            return None, None, args

    def list_commands(self, ctx):
        # prevent alphabetical ordering
        return list(self.commands)



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import inspect
from importlib.metadata import version

from pydantic import BaseModel

from . import models, service

__version__ = version("gavicore")

__all__ = [
    "__version__",
    "models",
    "service",
]


def _patch_models():
    for name, obj in inspect.getmembers(models, inspect.isclass):
        if (
            not name.startswith("_")
            and obj.__module__ == models.__name__
            and issubclass(obj, BaseModel)
        ):
            # Make model object render nicely in Jupyter notebooks
            setattr(obj, "_repr_json_", _repr_base_model_as_json)


def _repr_base_model_as_json(self: BaseModel):
    return self.model_dump(
        mode="json", by_alias=True, exclude_none=True, exclude_defaults=True
    ), dict(root=self.__class__.__name__ + " object:")


_patch_models()



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

import os
from collections.abc import Generator
from contextlib import contextmanager
from typing import Callable

from pydantic import BaseModel


# noinspection PyAttributeOutsideInit,PyUnresolvedReferences,PyPep8Naming
class BaseModelMixin:
    """
    Add to test classes to get the `assertBaseModelEqual` method.
    """

    def assertBaseModelEqual(self, a: BaseModel, b: BaseModel):
        """Assert that two `BaseModel`s are equal."""
        self.maxDiff = None
        self.assertIsInstance(a, BaseModel)  # type: ignore[attr-defined]
        self.assertIsInstance(b, BaseModel)  # type: ignore[attr-defined]
        self.assertIs(b.__class__, a.__class__)  # type: ignore[attr-defined]
        a_value = a.model_dump(
            mode="json", by_alias=True, exclude_none=True, exclude_defaults=True
        )
        b_value = b.model_dump(
            mode="json", by_alias=True, exclude_none=True, exclude_defaults=True
        )
        self.assertEqual(a_value, b_value)  # type: ignore[attr-defined]


@contextmanager
def set_env_cm(**new_env: str | None) -> Generator[dict[str, str | None]]:
    """Run the code in the block with a new environment `new_env`."""
    restore_env = set_env(**new_env)
    try:
        yield new_env
    finally:
        restore_env()


def set_env(**new_env: str | None) -> Callable[[], None]:
    """
    Set the new environment in `new_env` and return a no-arg
    function to restore the old environment.
    """
    old_env = {k: os.environ.get(k) for k in new_env.keys()}

    def restore_env():
        for ko, vo in old_env.items():
            if vo is not None:
                os.environ[ko] = vo
            elif ko in os.environ:
                del os.environ[ko]

    for kn, vn in new_env.items():
        if vn is not None:
            os.environ[kn] = vn
        elif kn in os.environ:
            del os.environ[kn]

    return restore_env



# generated by gen_server.py:
#   filename:  service.py:
#   timestamp: 2025-10-01T16:12:52.122326

from abc import ABC, abstractmethod

from .models import (
    Capabilities,
    ConformanceDeclaration,
    JobInfo,
    JobList,
    JobResults,
    ProcessDescription,
    ProcessList,
    ProcessRequest,
)


class Service(ABC):
    @abstractmethod
    async def get_capabilities(self, *args, **kwargs) -> Capabilities:
        """The landing page provides links to the:
          * The OpenAPI-definition (no fixed path),
          * The Conformance statements (path /conformance),
          * The processes metadata (path /processes),
          * The endpoint for job monitoring (path /jobs).

        For more information, see [Section 7.2](https://docs.ogc.org/is/18-062/18-062.html#sc_landing_page)."""

    @abstractmethod
    async def get_conformance(self, *args, **kwargs) -> ConformanceDeclaration:
        """A list of all conformance classes, specified in a standard, that the server conforms to.

        | Conformance class | URI |
        |-----------|-------|
        |Core|http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/core|
        |OGC Process Description|http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/ogc-process-description|
        |JSON|http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/json|
        |HTML|http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/html|
        |OpenAPI Specification 3.0|http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/oas30|
        |Job list|http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/job-list|
        |Callback|http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/callback|
        |Dismiss|http://www.opengis.net/spec/ogcapi-processes-1/1.0/conf/dismiss|

        For more information, see [Section 7.4](https://docs.ogc.org/is/18-062/18-062.html#sc_conformance_classes).
        """

    @abstractmethod
    async def get_processes(self, *args, **kwargs) -> ProcessList:
        """The list of processes contains a summary of each process the OGC API - Processes offers, including the link to a more detailed description of the process.

        For more information, see [Section 7.9](https://docs.ogc.org/is/18-062/18-062.html#sc_process_list).
        """

    @abstractmethod
    async def get_process(self, process_id: str, *args, **kwargs) -> ProcessDescription:
        """The process description contains information about inputs and outputs and a link to the execution-endpoint for the process. The Core does not mandate the use of a specific process description to specify the interface of a process. That said, the Core requirements class makes the following recommendation:

        Implementations SHOULD consider supporting the OGC process description.

        For more information, see [Section 7.10](https://docs.ogc.org/is/18-062/18-062.html#sc_process_description).
        """

    @abstractmethod
    async def execute_process(
        self, process_id: str, process_request: ProcessRequest, *args, **kwargs
    ) -> JobInfo:
        """Create a new job.

        For more information, see [Section 7.11](https://docs.ogc.org/is/18-062/18-062.html#sc_create_job).
        """

    @abstractmethod
    async def get_jobs(self, *args, **kwargs) -> JobList:
        """Lists available jobs.

        For more information, see [Section 11](https://docs.ogc.org/is/18-062/18-062.html#sc_job_list).
        """

    @abstractmethod
    async def get_job(self, job_id: str, *args, **kwargs) -> JobInfo:
        """Shows the status of a job.

        For more information, see [Section 7.12](https://docs.ogc.org/is/18-062/18-062.html#sc_retrieve_status_info).
        """

    @abstractmethod
    async def dismiss_job(self, job_id: str, *args, **kwargs) -> JobInfo:
        """Cancel a job execution and remove it from the jobs list.

        For more information, see [Section 13](https://docs.ogc.org/is/18-062/18-062.html#Dismiss).
        """

    @abstractmethod
    async def get_job_results(self, job_id: str, *args, **kwargs) -> JobResults:
        """Lists available results of a job. In case of a failure, lists errors instead.

        For more information, see [Section 7.13](https://docs.ogc.org/is/18-062/18-062.html#sc_retrieve_job_results).
        """



#  Copyright (c) 2025 by the Eozilla team and contributors
#  Permissions are hereby granted under the terms of the Apache 2.0 License:
#  https://opensource.org/license/apache-2-0.

from typing import Any


def flatten_obj(
    obj: Any, *, parent_key: str = "", sep: str = ".", flatten_lists: bool = False
) -> dict[str, Any]:
    return _flatten_obj(
        obj, parent_key=parent_key, sep=sep, flatten_lists=flatten_lists
    )


def _flatten_obj(
    obj: Any,
    parent_key: str = "",
    sep: str = ".",
    flatten_lists: bool = False,
) -> dict[str, Any]:
    """
    Flatten nested dicts/lists into a flat dict with sep-delimited keys.

    Args:
        obj: The nested structure (dict, list, or scalar).
        parent_key: Internal recursion prefix for keys.
        sep: Separator string for flattened keys.
        flatten_lists: If True, expand lists into indexed keys (default).
                       If False, keep lists as values (without flattening inside).

    Returns:
        A flat dictionary mapping sep-delimited keys to values.
    """
    items: dict[str, Any] = {}

    if isinstance(obj, dict) and len(obj) > 0:
        for k, v in obj.items():
            _update_flattened_items(k, v, parent_key, flatten_lists, sep, items)
    elif isinstance(obj, list) and len(obj) > 0 and flatten_lists:
        for i, v in enumerate(obj):
            if v is None:  # skip None placeholders
                continue
            _update_flattened_items(i, v, parent_key, flatten_lists, sep, items)
    elif parent_key:
        items[parent_key] = obj

    return items


def _update_flattened_items(
    k,
    v,
    parent_key: str,
    flatten_lists: bool,
    sep: str,
    items: dict[str, Any],
):
    new_key = f"{parent_key}{sep}{k}" if parent_key else str(k)
    flattened_v = _flatten_obj(
        v, parent_key=new_key, flatten_lists=flatten_lists, sep=sep
    )
    items.update(flattened_v)


def nest_dict(flatted_dict: dict[str, Any], *, sep: str = ".") -> dict[str, Any]:
    """Nest a flat dict with sep-delimited keys back into a dicts."""
    root = nest_obj(flatted_dict, sep=sep, root={})
    assert isinstance(root, dict)
    return root


def nest_obj(
    flatted_dict: dict[str, Any],
    *,
    sep: str = ".",
    root: dict[str, Any] | list[Any] | None = None,
) -> dict[str, Any] | list[Any] | None:
    """Nest a flat dict with sep-delimited keys back into dicts/lists."""
    # TODO: respect flatten_lists
    for k, v in flatted_dict.items():
        root = _nest_one(root, k, v, sep=sep)
    return root


def _nest_one(
    root: dict[str, Any] | list[Any] | None,
    name: str,
    value: Any,
    sep: str = ".",
) -> dict[str, Any] | list[Any] | None:
    keys = _parse_path(name, sep)

    current: Any = None
    prev: Any | None = None

    for i, key_or_index in enumerate(keys):
        if root is None:
            root = [] if isinstance(key_or_index, int) else {}
        if i == 0:
            current = root
        if current is None and prev is not None:
            current = prev[keys[i - 1]] = [] if isinstance(key_or_index, int) else {}
        prev = current
        current = _prepare_path(prev, key_or_index)

    assert prev is not None
    prev[keys[-1]] = value
    return root


def _prepare_path(current: dict[str, Any] | list[Any], key_or_index: str | int) -> Any:
    if isinstance(current, list):
        if not isinstance(key_or_index, int):
            raise TypeError(
                f"expected index of type int, got {type(key_or_index).__name__}"
            )
        index: int = key_or_index
        current_list: list[Any] = current
        while len(current_list) <= index:
            current_list.append(None)  # This is a Fill-value
        return current_list[index]
    elif isinstance(current, dict):
        if isinstance(key_or_index, int):
            raise TypeError(
                f"expected key of type str, got {type(key_or_index).__name__}"
            )
        key: str = key_or_index
        current_dict: dict[str, Any] = current
        if key not in current_dict:
            current_dict[key] = None  # This is a Fill-value
        return current_dict[key]
    raise TypeError(f"expected a list or a dict, got {type(current).__name__}")


def _parse_path(name: str, sep: str = ".") -> list[str | int]:
    keys: list[str] = name.split(sep)
    return [(int(key) if all(c.isdigit() for c in key) else key) for key in keys]
